/*-
 * insthook.c - Shared library hook that tracks installation sessions
 * 
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42, (c) Poul-Henning Kamp):
 * Hye-Shik Chang <perky@FreeBSD.org> wrote this file.  As long as you retain
 * this notice you can do whatever you want with this stuff. If we meet some
 * day, and you think this stuff is worth it, you can buy me a beer in return.
 *
 * Hye-Shik Chang
 * ----------------------------------------------------------------------------
 * $Id: insthook.c 497 2005-05-08 01:28:18Z perky $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdarg.h>
#include <string.h>
#include <err.h>
#include <errno.h>
#include <dlfcn.h>
#include <fcntl.h>
#include <libgen.h>
#include <sys/param.h>
#include <sysexits.h>

/* global variables */
static int insthook_initialized = 0;
static char *insthook_logfilename;

/* insthook static functions */
static int initialize(void);
static void writelog(const char *format, ...)
#ifdef __GNUC__
	__attribute__((format(printf, 1, 2)))
#endif
;

/* libc function pointers */
static int (*__creat)(const char *, mode_t);
static FILE *(*__fopen)(const char *, const char *);
static int (*__link)(const char *, const char *);
static int (*__mkdir)(const char *, mode_t);
static int (*__open)(const char *, int, ...);
static int (*__rename)(const char *, const char *);
static int (*__rmdir)(const char *);
static int (*__symlink)(const char *, const char *);
static int (*__truncate)(const char *, off_t length);
static int (*__unlink)(const char *);

/* macros */
#define D "\x01"			/* delimiter for log format */
#define L "\n"				/* line delimiter for log format */
#define ERRX(why) errx(why, strerror(errno))	/* show error and exit */
#define INITIALIZE() do {		/* check whether initialized */	\
	if (!insthook_initialized && initialize() == -1)		\
		ERRX(EX_SOFTWARE);					\
} while (0)


static void
writelog(const char *format, ...)
{
	va_list ap;
	char logstr[BUFSIZ];
	int loglen, fd, r;

	va_start(ap, format);
	loglen = vsnprintf(logstr, BUFSIZ, format, ap);
	va_end(ap);

	if (loglen < 0)
		ERRX(EX_OSERR);

	fd = __open(insthook_logfilename, O_WRONLY | O_CREAT | O_APPEND, 0666);
	if (fd == -1)
		ERRX(EX_CANTCREAT);

	r = write(fd, logstr, loglen);
	if (r == -1)
		ERRX(EX_IOERR);

	(void)close(fd);
}

static int
initialize(void)
{
	char *path;

#define LIBC_FUN(name) do {					\
	__##name = dlsym(RTLD_NEXT, #name);			\
	if (__##name == NULL)					\
		return -1;					\
} while (0)

	LIBC_FUN(creat);
	LIBC_FUN(fopen);
	LIBC_FUN(link);
	LIBC_FUN(mkdir);
	LIBC_FUN(open);
	LIBC_FUN(rename);
	LIBC_FUN(rmdir);
	LIBC_FUN(symlink);
	LIBC_FUN(truncate);
	LIBC_FUN(unlink);

#undef LIBC_FUN

	path = getenv("INSTHOOK_LOGFILE");
	if (path == NULL)
		return -1;
	insthook_logfilename = strdup(path);

	writelog("%d" D "newproc" D "%s" L, getpid(), getprogname());
	insthook_initialized = 1;

	return 0;
}

static void
canonicalize_full(const char *path, char *resolved_path)
{
	char *r;

	r = realpath(path, resolved_path);
	if (r == NULL && path[0] != '/') {
		getcwd(resolved_path, PATH_MAX - 2);
		strcat(resolved_path, "/");
		strncat(resolved_path, path, PATH_MAX - 1);
	}
}

static void
canonicalize_dirname(const char *path, char *resolved_path)
{
	char dirpart[PATH_MAX];
	char *r;

	r = dirname(path);
	if (r == NULL)
		ERRX(EX_OSERR);
	else
		strcpy(dirpart, r);

	canonicalize_full(dirpart, resolved_path);

	r = basename(path);
	if (r == NULL)
		ERRX(EX_OSERR);

	if (strlen(resolved_path) + strlen(r) + 2 > PATH_MAX)
		ERRX(EX_DATAERR);

	strcat(resolved_path, "/");
	strcat(resolved_path, r);
	resolved_path[PATH_MAX - 1] = 0;
}

int
creat(const char *path, mode_t mode)
{
	int ret;
	char canon_path[PATH_MAX];

	INITIALIZE();
	canonicalize_full(path, canon_path);

	ret = __creat(path, mode);
	writelog("%d" D "creat" D "%d" D "%s" L, getpid(),
			ret != -1 ? 0 : errno, canon_path);
	return ret;
}

FILE *
fopen(const char * __restrict path, const char * __restrict mode)
{
	FILE *ret;
	char canon_path[PATH_MAX];

	INITIALIZE();
	canonicalize_full(path, canon_path);

	ret = __fopen(path, mode);

	if (mode[0] == 'w' || mode[0] == 'a' || mode[1] == '+' /* r+ */)
		writelog("%d" D "fopen" D "%d" D "%s" L, getpid(),
				ret == NULL ? errno : 0, canon_path);

	return ret;
}

int
link(const char *name1, const char *name2)
{
	int ret;
	char canon_name1[PATH_MAX], canon_name2[PATH_MAX];

	INITIALIZE();
	canonicalize_full(name1, canon_name1);
	canonicalize_dirname(name2, canon_name2);

	ret = __link(name1, name2);
	writelog("%d" D "link" D "%d" D "%s" D "%s" L, getpid(),
			ret != -1 ? 0 : errno, canon_name1, canon_name2);
	return ret;
}

int
mkdir(const char *path, mode_t mode)
{
	int ret;
	char canon_path[PATH_MAX];

	INITIALIZE();
	canonicalize_dirname(path, canon_path);

	ret = __mkdir(path, mode);
	writelog("%d" D "mkdir" D "%d" D "%s" L, getpid(),
			ret == 0 ? 0 : errno, canon_path);
	return ret;
}

int
open(const char *path, int flags, ...)
{
	int ret;
	va_list ap;
	mode_t mode;
	char canon_path[PATH_MAX];

	INITIALIZE();
	canonicalize_full(path, canon_path);

	va_start(ap, flags);
	mode = (int)va_arg(ap, int);
	va_end(ap);

	ret = __open(path, flags, mode);

	if (flags & (O_WRONLY | O_RDWR))
		writelog("%d" D "open" D "%d" D "%s" L, getpid(),
				ret != -1 ? 0 : errno, canon_path);
	return ret;
}

int
rename(const char *from, const char *to)
{
	int ret;
	char canon_from[PATH_MAX], canon_to[PATH_MAX];

	INITIALIZE();
	canonicalize_dirname(from, canon_from);
	canonicalize_dirname(to, canon_to);

	ret = __rename(from, to);
	writelog("%d" D "rename" D "%d" D "%s" D "%s" L, getpid(),
			ret != -1 ? 0 : errno, canon_from, canon_to);
	return ret;
}

int
rmdir(const char *path)
{
	int ret;
	char canon_path[PATH_MAX];

	INITIALIZE();
	canonicalize_dirname(path, canon_path);

	ret = __rmdir(path);
	writelog("%d" D "rmdir" D "%d" D "%s" L, getpid(),
			ret != -1 ? 0 : errno, canon_path);
	return ret;
}

int
symlink(const char *name1, const char *name2)
{
	int ret;
	char canon_name1[PATH_MAX], canon_name2[PATH_MAX];

	INITIALIZE();
	canonicalize_dirname(name1, canon_name1);
	canonicalize_dirname(name2, canon_name2);

	ret = __symlink(name1, name2);
	writelog("%d" D "symlink" D "%d" D "%s" D "%s" L, getpid(),
			ret != -1 ? 0 : errno, canon_name1, canon_name2);
	return ret;
}

int
truncate(const char *path, off_t length)
{
	int ret;
	char canon_path[PATH_MAX];

	INITIALIZE();
	canonicalize_full(path, canon_path);

	ret = __truncate(path, length);
	writelog("%d" D "truncate" D "%d" D "%s" D "%lld" L, getpid(),
			ret != -1 ? 0 : errno, canon_path, (long long)length);
	return ret;
}

int
unlink(const char *path)
{
	int ret;
	char canon_path[PATH_MAX];

	INITIALIZE();
	canonicalize_dirname(path, canon_path);

	ret = __unlink(path);
	writelog("%d" D "unlink" D "%d" D "%s" L, getpid(),
			ret != -1 ? 0 : errno, canon_path);
	return ret;
}
